//----------------------------------------------------------------------------//
//                                                                            //
//                                                                            //
//                                                                            //
//                              CORE D3 FUNCTIONS                             //
//                                                                            //
//                                                                            //
//                                                                            //
//----------------------------------------------------------------------------//


//-------------------------------------------------------------------//
//                                                                   //
//                          SVG EXPORT FUNCTIONS                     //
//                                                                   //
//-------------------------------------------------------------------//
// for each button with class .export_button applied the onclick event
// export_chart_as_svg
function setup_export_buttons() {
  d3.selectAll(".export_button").on("click", export_chart_as_svg);
}

function export_chart_as_svg(){
    // something like feature_selection_chart
    var this_id = d3.select(this).node().id
    // feature_selection
    var chart_name = this_id.split("_chart")[0]
    // The corresponding name of the sv
    var svg_id = chart_name + "_svg"
    try {
        var isFileSaverSupported = !!new Blob();
    } catch (e) {
        alert("blob not supported");
    }

    // Get the SVG HTML
    var html = d3.select("svg#" + svg_id)
        .attr("version", 1.1)
        .attr("xmlns", "http://www.w3.org/2000/svg")
        .node().outerHTML


    // Make the blob thing and save it
    var blob = new Blob([html], {type: "image/svg+xml"});
    saveAs(blob, chart_name + ".svg");
};


//-------------------------------------------------------------------//
//                                                                   //
//                     USE-CASE SPECIFIC FUNCTIONS                   //
//                                                                   //
//-------------------------------------------------------------------//
function make_all_charts(models) {
  make_pca_plot(models)
  make_outlier_plot(models)
  make_roc_plot(models)
  make_pr_plot(models)
  make_vi_plot(models)
  make_oob_plot(models)
  make_fs_plot(models)

  // Cell widths of a table with stick header needed to be updated to work.
  // Just another stupid quirk of JS
  update_table_header_width(charts_config.table_id)
  update_table_header_width("oasis_classification_feature_importance_table")
  // Update which xls file is to be downloaded
  update_download_attribute("oasis_classification_feature_importance_table_download_button", "classification.xls")
  // Update which RData file is to be downloaded
  update_download_attribute("model_download_link", "rf_model.RData")
}


//-------------------------------------------------------------------//
//                            DRAG BOX                               //
//-------------------------------------------------------------------//
function drag_box_function(d) {
  var old_x = parseFloat(d3.select(this).style("left"), 10)
  var new_x = d3.mouse(this)[0]
  var old_y = parseFloat(d3.select(this).style("top"), 10)
  var new_y = d3.mouse(this)[1]

  d3.select(this).raise()
  d3.select(this).style("left", (old_x + new_x)+"px")
  d3.select(this).style("top", (old_y + new_y)+"px")
}


function setup_dragbox() {
  d3.select("div.drag-box").call(
    d3.drag().on("drag", drag_box_function)
  )

  d3.select("i").on("click", function() {
    var cur_disp = d3.select("span.icon_click_toggle").style("display")
    if (cur_disp == "none") {
      d3.select("span.icon_click_toggle").style("display", "block")
    } else {
      d3.select("span.icon_click_toggle").style("display", "none")
    }
  })
}
//-------------------------------------------------------------------//
//                        RETRIEVE JSON DATA                         //
//-------------------------------------------------------------------//
// Gather model data from json files located under the corresponding files
// provided in the charts_config.files object
function loadModels() {
  var models = {"balanced": {},"unbalanced": {}}
  // Load balanced all variable model
  models.balanced["all_variables"] = transformOldJSON(ajax_json(charts_config.files.balanced["all_variables"]))
  // Load balanced optimized variable model
  models.balanced["optimized_variables"] = transformOldJSON(ajax_json(charts_config.files.balanced["optimized_variables"]))
  // Load unbalanced all variable model
  models.unbalanced["all_variables"] = transformOldJSON(ajax_json(charts_config.files.unbalanced["all_variables"]))
  // Load unbalanced optimized variable model
  models.unbalanced["optimized_variables"] = transformOldJSON(ajax_json(charts_config.files.unbalanced["optimized_variables"]))
  return models
}

//-------------------------------------------------------------------//
//                     RETRIEVE SELECTED DATA                        //
//-------------------------------------------------------------------//
// Return the model of the current document state from charts_config
function extractModel(models) {
  var model;
  if (charts_config["document_state"].balanced) {
    if (charts_config["document_state"].optimized) {
      model = models.balanced["optimized_variables"]
    } else {
      model = models.balanced["all_variables"]
    }
  } else {
    if (charts_config["document_state"].optimized) {
      model = models.unbalanced["optimized_variables"]
    } else {
      model = models.unbalanced["all_variables"]
    }
  }
  return model
}

//-------------------------------------------------------------------//
//                       UPDATE SELECTED DATA                        //
//-------------------------------------------------------------------//
function update_charts(value, models) {
  if (value == "balanced") {
    if (charts_config["document_state"].balanced) {
      charts_config["document_state"].balanced = false
    } else {
      charts_config["document_state"].balanced = true
    }
  } else {
    if (charts_config["document_state"].optimized) {
      charts_config["document_state"].optimized = false
    } else {
      charts_config["document_state"].optimized = true
    }
  }

  populate_table()
  make_all_charts(models)
  classification_results_table()

  d3.selectAll('.target_div').remove()
}


//-------------------------------------------------------------------//
//                  POPULATE TABLE OF SELECTED DATA                  //
//-------------------------------------------------------------------//
function populate_table() {
  var table = d3.select("#" + charts_config.table_id)
  var table_head = table.select("thead")
  var table_body = table.select("tbody")
  var stats = extractModel(models).statistics
  var stats_key = keys(stats)

  // Create header if need be
  if (table_head.empty()) {
    table_head = table.append("thead")
    table_head_row = table_head.append("tr")
    table_head_row.append("th").node().innerHTML =  "MEASURE"
    table_head_row.append("th").node().innerHTML = "VALUE"
  }

  // Create body if need be
  if (table_body.empty()) {
    table_body = table.append("tbody")
  }


  var previous_rows = table_body.selectAll("tr").size()
  var current_rows = stats_key.length

  // Add or remove table rows as need be
  if (previous_rows < current_rows) {
    var rows = table_body.selectAll("tr")
                    .data(stats_key)
                    .enter()
                    .append("tr")
    rows.each(function(d, i) {
      var table_row = d3.select(this)
      var table_key = table_row.append("td").attr("class", "table_key")
      var table_val = table_row.append("td").attr("class", "table_val")
    })
  } else if (previous_rows > current_rows) {
    table_body.selectAll("tr").data(stats_key)
         .exit().remove()
  }

  // Populate values
  rows = table_body.selectAll("tr").each(function(d, i) {
    var table_row = d3.select(this)
    var table_key = table_row.select("td.table_key")
    var table_val = table_row.select("td.table_val")
    var measure = d
    var value = stats[measure]
    // console.log(measure, value)
    if (measure == "O.O.B" || measure == "CV error") {
      value += "%"
    } else if (parseFloat(value) != NaN) {
      var [tens, decimal] = parseFloat(value).toString().split(".")
      // console.log(tens, decimal)
    }

    table_key.node().innerHTML = measure
    table_val.node().innerHTML = value

  })


  update_table_header_width(charts_config.table_id)

}

//-------------------------------------------------------------------//
//                                                                   //
//                         HELPER FUNCTIONS                          //
//                                                                   //
//-------------------------------------------------------------------//
function parseNumber(number) {
  var number_string = number.toString()
  if (number_string.includes("%")) {
    number_string = number_string.slice(0, number_string.length - 1)
    if (number_string.length == 3) {
      return 1.0
    } else {
      return parseFloat("." + number_string)
    }
  } else if (number_string.includes(".")) {
      return parseFloat(number_string)
  } else {
    return parseInt(number_string)
  }
}

function typeofNumber(number) {
  number = number.toString()
  if (number.includes("%")) {
    return "percent"
  } else if (number.includes(".")) {
    return "float"
  } else if (number.includes("e") || number.includes("+")) {
    return "scientific"
  } else {
    return "integer"
  }
}

function extract_all_groups(data, groupKey) {
  var groups = []
  for (var i = 0; i < data.length; i++) {
    if (groups.indexOf(data[i][groupKey]) == -1) {
      groups.push(data[i][groupKey])
    }
  }
  return groups
}

//-------------------------------------------------------------------//
//                                                                   //
//                            MAKE AXES                              //
//                                                                   //
//-------------------------------------------------------------------//
function calculate_space_needed_by_axes(chart_group, data_extent, margins) {
  // Temporary SVG group element to calculate size of dummy axes then be removed
  var temp = chart_group.append("g").attr("class", "temp")

  // Dummy scales. Use max and min to get accurate measurement for length of ticks
  var x_scale = d3.scaleLinear().domain([data_extent.min.x, data_extent.max.x]).range([0, 1])
  var y_scale = d3.scaleLinear().domain([data_extent.min.y, data_extent.max.y]).range([0, 1])

  // Make the axes
  var x_axis_scaled = d3.axisBottom().scale(x_scale).tickSize(charts_config.plot_attributes.axes.ticks.size).ticks(6)
  var y_axis_scaled = d3.axisLeft().scale(y_scale).tickSize(charts_config.plot_attributes.axes.ticks.size).ticks(6)

  // Place and maintain the axis as an SVG object
  var x_axis= temp.append("g").attr("class", "temp").call(x_axis_scaled).attr("font-size", charts_config.plot_attributes.axes.size).attr("font-family", charts_config.plot_attributes.axes.family)
  var y_axis= temp.append("g").attr("class", "temp").call(y_axis_scaled).attr("font-size", charts_config.plot_attributes.axes.size).attr("font-family", charts_config.plot_attributes.axes.family)

  // Add the axes labels. y-axis is rotated
  var x_axis_label = temp.append("text").text("temp").attr("font-family", charts_config.plot_attributes.axes.labels.family).attr("font-size", charts_config.plot_attributes.axes.labels.size)
  var y_axis_label = temp.append("text").text("temp").attr("font-family", charts_config.plot_attributes.axes.labels.family).attr("font-size", charts_config.plot_attributes.axes.labels.size).attr("transform", "rotate(-90)")

  // Retrieve the rectangle encapsulating the text labels and the axes
  var x_axis_label_box = x_axis_label.node().getBBox()
  var y_axis_label_box = y_axis_label.node().getBBox()
  var x_axis_box = x_axis.node().getBBox()
  var y_axis_box = y_axis.node().getBBox()


  // Calculate the total space consumed by these items
  var x_axis_space_consumed = x_axis_label_box.height + x_axis_box.height + margins.axes.label_space
  var y_axis_space_consumed = y_axis_label_box.height + y_axis_box.width + margins.axes.label_space

  // Remove these SVG elements
  temp.remove()
  return {"x": x_axis_space_consumed, "y": y_axis_space_consumed}
}

function make_axes(chart_group, x_scale, y_scale, canvas, margins, maximum_drawing_space, x_label, y_label, custom_ticks) {
  if (custom_ticks == undefined) {
    custom_ticks = false
  }


  var axes, x_axis_label, y_axis_label, x_axis, y_axis
  // If first call make all groups needed
  if (chart_group.select("g.axes").empty()) {
    axes = chart_group.append("g").attr("class", "axes")
    x_axis_label = axes.append("text").attr("class", "x_axis_label")
    y_axis_label = axes.append("text").attr("class", "y_axis_label")
    x_axis= axes.append("g").attr("class", "x_axis")
    y_axis= axes.append("g").attr("class", "y_axis")
  }

  axes = chart_group.select("g.axes")
  var x_axis_scaled = d3.axisBottom().scale(x_scale).tickSize(charts_config.plot_attributes.axes.ticks.size).ticks(6)
  var y_axis_scaled = d3.axisLeft().scale(y_scale).tickSize(charts_config.plot_attributes.axes.ticks.size).ticks(6)
  var center = {"x": margins.x.left + margins.axes.y + maximum_drawing_space.x / 2,"y": margins.y.top +  margins.buttons + margins.title + maximum_drawing_space.y / 2}

  if (custom_ticks!=false) {
    x_axis_scaled.tickFormat(function (d, i) {return custom_ticks[i]}).ticks(custom_ticks.length)
  }



  x_axis_label = axes.select("text.x_axis_label").text(x_label).attr("font-family", charts_config.plot_attributes.axes.labels.family).attr("font-size", charts_config.plot_attributes.axes.labels.size).attr("transform", "translate(" + center.x + "," + (canvas.y - margins.y.bottom)+")").attr("text-anchor", "middle")
  y_axis_label = axes.select("text.y_axis_label").text(y_label).attr("font-family", charts_config.plot_attributes.axes.labels.family).attr("font-size", charts_config.plot_attributes.axes.labels.size).attr("transform", "translate(" + margins.x.left + "," + center.y + ") rotate(-90)").attr("text-anchor", "middle")

  var x_axis_label_box = x_axis_label.node().getBBox()
  var y_axis_label_box = y_axis_label.node().getBBox()

  var x_axis_y = margins.axes.x + margins.y.bottom
  var y_axis_x = margins.axes.y + margins.x.left
  x_axis= axes.select("g.x_axis").transition().duration(500).call(x_axis_scaled)
                 .attr("transform", "translate(" + (margins.x.left + margins.axes.y) + "," + (canvas.y - x_axis_y)+")")
                 .attr("font-size", charts_config.plot_attributes.axes.size)
                 .attr("font-family", charts_config.plot_attributes.axes.family)



  y_axis= axes.select("g.y_axis").transition().duration(500).call(y_axis_scaled)
                 .attr("font-size", charts_config.plot_attributes.axes.size)
                 .attr("font-family", charts_config.plot_attributes.axes.family)
                 .attr("transform", "translate(" + y_axis_x + "," + (center.y - maximum_drawing_space.y / 2)+")")

}


//-------------------------------------------------------------------//
//                                                                   //
//                         MAKE SVG BUTTONS                          //
//                                                                   //
//-------------------------------------------------------------------//
// Makes radio buttons for changing chart
function make_buttons(chart_group, margins, canvas, maximum_drawing_space, button_array, button_click_function) {
    var radius = charts_config.plot_attributes.buttons.size / 2 // radius is half button font size
    var x = margins.axes.y + margins.x.left + radius // starting x locations
    var y = radius + margins.y.top

    var button_group
    if (chart_group.selectAll("g.buttons_group").empty()) {
      chart_group.append("g") // group for all buttons
                .attr("class", "buttons_group")
                .selectAll("g.buttons_group")
                .data(button_array)
                .enter()
                .append("g") // group for each button (circle and text)
                .attr("class", "button_group")

      var button_group = chart_group.selectAll("g.button_group")
      // add the circles
      button_group.append("circle")
                 .attr("r", radius)
                 .attr("class", "button_circle")
                 .attr("class", function(d, i) {return "button_circle_number" + i})
                 .on("click", button_click_function)
      // add the text
      button_group.append("text")
                 .attr("class", "button_text")
                 .text(function(d) {return d})
                 .attr("class", function(d, i) {return "button_text_number" + i})
                 .on("click", button_click_function)
    }
    button_group = chart_group.selectAll("g.button_group")
    // position the circles and text
    button_group.each(function(d, i) {
      var current_group = d3.select(this)
      var current_circle = current_group.select("circle")
      var current_text = current_group.select("text")
      if (i == 0) {
        current_circle.attr("cx", x)
                     .attr("cy", y)
                     .attr("fill", charts_config.plot_attributes.buttons.fill.not_selected)
                     .attr("stroke", charts_config.plot_attributes.buttons.stroke)

        current_text.attr("x", x + radius * 2)
                   .attr("y", y + radius / 2)
                   .text(function(d, i) {return d})
                   .attr("font-family", charts_config.plot_attributes.buttons.family)
                   .attr("font-size", charts_config.plot_attributes.buttons.size)
        margins.buttons = current_text.node().getBBox().y + current_text.node().getBBox().height + radius
      } else {
        var previous_text = button_group.select(".button_text_number" + (i-1))
        var previous_circle = button_group.select(".button_circle_number" + (i-1))
        var previous_text_box = previous_text.node().getBBox()
        var previous_circle_box = previous_circle.node().getBBox()

        var cx = previous_text_box.x + previous_text_box.width + radius * 2
        var cy = previous_circle_box.y + radius
        current_circle.attr("cx", cx)
                     .attr("cy", cy)
                     .attr("fill", charts_config.plot_attributes.buttons.fill["not_selected"])
                     .attr("stroke", charts_config.plot_attributes.buttons.stroke)
        current_text.attr("x", cx + radius * 2)
                   .attr("y", cy + radius / 2)
                   .text(function(d, i) {return d})
                   .attr("font-family", charts_config.plot_attributes.buttons.family)
                   .attr("font-size", charts_config.plot_attributes.buttons.size)

        var far_right_point = current_text.node().getBBox().x + current_text.node().getBBox().width
        var farthest_point_to_draw = canvas.x - margins.x.right

        if (far_right_point > farthest_point_to_draw) { //
           cx = x
           cy = previous_circle_box.y + radius * 4
           current_circle.attr("cx", cx)
                        .attr("cy", cy)
           current_text.attr("x", cx + radius * 2)
                      .attr("y", cy + radius / 2)
        }
        margins.buttons = current_text.node().getBBox().y + current_text.node().getBBox().height + radius
      }
    })
}


function color_buttons(chart_group, state) {
  chart_group.select("g.buttons_group").selectAll("circle").attr("fill", function (d, i) {
    if (d == state) {
      return charts_config.plot_attributes.buttons.fill.selected
    }
    return charts_config.plot_attributes.buttons.fill.not_selected
  })
}





//-------------------------------------------------------------------//
//                                                                   //
//                           MAKE MARGINS                            //
//                                                                   //
//-------------------------------------------------------------------//
function make_margins(chart_group, canvas, data_extent, buttons) {
  if (buttons == undefined) {
    buttons = true
  }
  var margins = {
    "x": {
      "left": canvas.x * 0.04,
      "right": canvas.x * 0.04
    },
    "y": {
      "top": canvas.x * 0.04,
      "bottom": canvas.x * 0.04
    },
    "title": charts_config.plot_attributes.title.size * 2, // space consumed by title
    "buttons": charts_config.plot_attributes.buttons.size * 2, // space consumed by buttons
    "axes": {
      "x": charts_config.plot_attributes.axes.size * 4, // space consumed by x axes
      "y": charts_config.plot_attributes.axes.size * 4, // space consumed by y axes
      "label_space": 10 // space between axis label and axis, included in axes.x / axes.y respectively
    }
  }

  // Update axes margins based off space used by their rendered SVG elements
  axes_margins = calculate_space_needed_by_axes(chart_group, data_extent, margins)
  margins.axes.x = axes_margins.x
  margins.axes.y = axes_margins.y

  if (!buttons) {margins.buttons = 0}
  return margins
}

function calculate_maximum_drawing_space(canvas, margins) {
  var maximum_drawing_space = {
    "x": canvas.x - margins.x.left - margins.x.right - margins.axes.y,
    "y": canvas.y - margins.y.top - margins.y.bottom - margins.axes.x - margins.title - margins.buttons
  };
  return maximum_drawing_space
}

function extract_canvas_from_svg(svg) {
  var width = svg.attr("width")
  var height = svg.attr("height")
  var canvas = {
    "x": parseNumber(width),
    "y": parseNumber(height)
  }
  return canvas
}





//-------------------------------------------------------------------//
//                                                                   //
//                             MAKE SVG                              //
//                                                                   //
//-------------------------------------------------------------------//
function make_chart_svg(plot) {
  var section = d3.select("#" + plot + "_chart")

  var chart_width = parseNumber(charts_config.svg.width)
  var chart_height = parseNumber(charts_config.svg.height)

  var chart_width_string, chart_height_string

  if (typeofNumber(chart_width) == "float") {
    chart_width_string = (chart_width * section.node().clientWidth ) + "px"
  } else {
    chart_width_string = chart_width + "px"
  }

  if (typeofNumber(chart_height) == "float") {
    chart_height_string = (chart_height * section.node().clientWidth ) + "px"
  } else {
    chart_height_string = chart_height + "px"
  }

  var svg
  if (section.select("#" +  plot + "_svg").empty()) {
    svg = section.insert("svg", ":nth-child(3)")
  } else {
    svg = section.select("#" +  plot + "_svg")
  }
  svg.attr("perserveAspectRatio", "xMinYMid meet")
         .classed("svg-content-responsive", true)
         .attr("id", plot + "_svg")
         .attr("width", chart_width_string)
         .attr("height", chart_height_string)
  return svg
}


//-------------------------------------------------------------------//
//                                                                   //
//                            MAKE TITLE                             //
//                                                                   //
//-------------------------------------------------------------------//
function make_title(chart_group, text_array, margins, canvas, maximum_drawing_space) {
  // Does chart title already exist?
  if (!chart_group.select("g.chart_title").empty()) {
    // Yes. Clear Title
    chart_group.select("g.chart_title").remove()
    // Reset Margins
    margins.title = charts_config.plot_attributes.title.size * 2
  }

  // Construct full title
  var full_title = text_array.join(" ")

  // Store title lines
  var title_lines = []

  // Max line length
  var characters_per_line = (canvas.x - margins.x.right) / (charts_config.plot_attributes.title.size / 2)

  while (full_title.length > 0) {
    var slice_position = characters_per_line - 1

    var line_slice = full_title.slice(0, slice_position)

    var last_space = line_slice.lastIndexOf(" ")

    // space is first character, drop it
    if (line_slice[0] == " ") {
      full_title = full_title.slice(1, full_title.length)
      line_slice = full_title.slice(0, slice_position)
      last_space = line_slice.lastIndexOf(" ")
    }

    if (full_title[slice_position + 1] != " " & slice_position < full_title.length) {
      // the leading character of next splice is not a space (e.g. breaks a word)
      // and there is more in the title to come
      if (last_space == -1) { // no spaces in this line, we have broken a word
        line_slice = full_title.slice(0, slice_position - 1) + "-"
        slice_position -= 1
      } else { // there is a space, truncate to that space
        slice_position = last_space
        line_slice = full_title.slice(0, slice_position)
      }
      last_space = line_slice.lastIndexOf(" ")
    } else if (slice_position < full_title.length & last_space < line_slice.length) {
      // last word is split, so add a hypen
      line_slice = full_title.slice(0, slice_position - 1)
      slice_position -= 1
      last_space = line_slice.lastIndexOf(" ")

      // if the word is a two letter word, e.g. the last letter in the string is
      // the first letter of the two letter word, then that letter is droped for
      // a hypen before a space. That makes no sense, so drop the entire 2 letter word
      if (last_space == slice_position -  1) {
        // if space is last character drop it
        line_slice = full_title.slice(0, slice_position - 1)
        slice_position -= 1
      } else {
        line_slice += "-"
        slice_position -= 1
      }
      last_space = line_slice.lastIndexOf(" ")
    }

    if (last_space == slice_position) {
      // if space is last character drop it
      line_slice = full_title.slice(0, slice_position - 1)
      slice_position -= 1
    }

    title_lines.push(line_slice)
    full_title = full_title.slice(slice_position, full_title.length)
  }

  var chart_title = chart_group.append("g").attr("class", "chart_title")
  for (var i = 0; i < title_lines.length; i++) {
    chart_title.append("text")
    .attr("class", "chartTitle")
    .text(title_lines[i])
    .attr("text-anchor", "middle")
    .attr("font-size", charts_config.plot_attributes.title.size)
    .attr("font-family", charts_config.plot_attributes.title.family)
    .attr("x", margins.axes.y + margins.x.left + maximum_drawing_space.x / 2)
    .attr("y", margins.y.top + margins.buttons + (charts_config.plot_attributes.title.size * (i) + charts_config.plot_attributes.title.size / 2))
    margins.title = chart_title.node().getBBox().height + charts_config.plot_attributes.title.size
    maximum_drawing_space.y = canvas.y - margins.y.top - margins.y.bottom - margins.axes.x - margins.title - margins.buttons
  }
}





//-------------------------------------------------------------------//
//                                                                   //
//                          MAKE TOOLTIP                             //
//                                                                   //
//-------------------------------------------------------------------//
function make_tooltip(obj, text_array, chart_group, canvas, margins) {
  // get the x and y coordinates of the object to apply tooltip too
  var x = obj.node().transform.baseVal.consolidate().matrix.e
  var y = obj.node().transform.baseVal.consolidate().matrix.f

  // for convenience
  var tooltip_config = charts_config.plot_attributes.tooltip
  var sep = tooltip_config.default_seperation_from_object

  // Add the tooltip to the chart and as a child - the text group
  var tooltip = chart_group.append("g").attr("class", "tooltip_group")
  var text_group = tooltip.append("g").attr("class", "tooltip_text_group")

  // Add text in reverse order placing low to high
  text_array.reverse()
  for (var i = 0; i < text_array.length; i++) {
    var text = text_group.append("text")
                        .text(text_array[i])
                        .attr("font-size", tooltip_config.size)
                        .attr("text-anchor", "middle")
                        .attr("x", x)
                        .attr("y", y - sep - tooltip_config.curve - (i * tooltip_config.size))
                        .attr("class", "tooltipText")
                        .attr("fill", tooltip_config.emphasis)
                        .attr("font-family", tooltip_config.family)
                        .attr("font-weight", "normal")

    // The first line in the tooltip gets different coloration
    if (i < text_array.length - 1) {
        text.attr("fill", tooltip_config.text)
            .attr("font-family", tooltip_config.family)
            .attr("font-weight", "normal")
    }
  }

  // Make the bubble around the text
  var bubble = tooltip.append("path")
                      .attr("d", make_tooltip_bubble(text_group))
                      .attr("fill", tooltip_config.fill)
                      .attr("stroke", tooltip_config.stroke)
                      .attr("opacity", tooltip_config.opacity)

  // Text goes in front of the box
  text_group.raise()

  // Get the bounding box of the bubble
  var bubble_box = bubble.node().getBBox()

  // Calculate the limits to contain the tooltip
  var limits = {
    "top": margins.y.top + margins.buttons + margins.title,
    "bottom": canvas.y - margins.axes.x - margins.y.bottom,
    "left": margins.x.left + margins.axes.y,
    "right": canvas.x - margins.x.right
  }

  // Get the boundaries of the object
  var object_boundaries = {
    "left": x,
    "right": x + obj.node().getBBox().width,
    "top": y,
    "bottom": y + obj.node().getBBox().height
  }

  // Calculate putative tooltip placements
  var tooltip_placements = {
    "upper_left": {"x": object_boundaries.left - bubble_box.width - sep, "y":object_boundaries.top - bubble_box.height - sep},
    "upper_middle": {"x": object_boundaries.left - bubble_box.width - sep + bubble_box.width / 2, "y":object_boundaries.top - bubble_box.height - sep},
    "upper_right": {"x":object_boundaries.right + sep, "y": object_boundaries.top - bubble_box.height - sep},
    "lower_left": {"x":object_boundaries.left - bubble_box.width - sep, "y": object_boundaries.bottom + sep},
    "lower_middle": {"x": object_boundaries.left - bubble_box.width - sep + bubble_box.width / 2, "y": object_boundaries.bottom + sep},
    "lower_right": {"x":object_boundaries.right + sep, "y":object_boundaries.bottom + sep}
  }


  // Figure out which placements fall within the limits
  var valid_placments = {
    "upper_left": {"x": tooltip_placements.upper_left.x > limits.left, "y":  tooltip_placements.upper_left.y > limits.top},
    "upper_middle": {"x": tooltip_placements.upper_middle.x > limits.left, "y": tooltip_placements.upper_left.y > limits.top},
    "upper_right": {"x": tooltip_placements.upper_right.x + bubble_box.width < limits.right, "y": tooltip_placements.upper_right.y > limits.top},
    "lower_left": {"x": tooltip_placements.lower_left.x > limits.left, "y": tooltip_placements.lower_left.y + bubble_box.height < limits.bottom},
    "lower_middle": {"x": tooltip_placements.lower_middle.x > limits.left, "y": tooltip_placements.lower_left.y + bubble_box.height < limits.bottom},
    "lower_right": {"x": tooltip_placements.lower_right.x + bubble_box.width < limits.right, "y": tooltip_placements.lower_right.y + bubble_box.height < limits.bottom}
  }

  var placements = keys(valid_placments)
  var tooltip_placement
  for (var i = 0; i < placements.length; i++) {
    var current_placement = valid_placments[placements[i]]
    if (current_placement.x && current_placement.y) {
      tooltip_placement = tooltip_placements[placements[i]]
      break
    }
  }

  if (tooltip_placement == undefined) {
    tooltip_placement = tooltip_placements.upper_right
  }


  // Reposition the tooltip to its correct location
  tooltip.attr("transform", "translate("+(tooltip_placement.x - bubble_box.x)+","+(tooltip_placement.y - bubble_box.y)+")")
}

function transform_values(selection) {
  return {
    "x": selection.node().transform.baseVal.consolidate().matrix.e,
    "y": selection.node().transform.baseVal.consolidate().matrix.f
  }
}

function make_tooltip_bubble(text_group) {
  var textBox = text_group.node().getBBox()
  var x = textBox.x
  var y = textBox.y
  var width = textBox.width
  var height = textBox.height

  var point = charts_config.plot_attributes.tooltip.point
  var curve = charts_config.plot_attributes.tooltip.curve

  // Start at bottom center and work around left - up - right - down - close
  d = "M " + (x + width / 2) + " " + (y + height + curve)
  // go left
  d += "l -" + (width / 2) + " 0 "
  // curve left and up
  d += "q -" + curve + " 0 -" + curve + " -" + curve + " "
  // go up
  d += "l 0 -" + (height) + " "
  // curve up and right
  d += "q 0 -" + curve + " " + curve + " -" + curve + " "
  // go right
  d += "l " + (width) + " 0 "
  // curve right and down
  d += "q " + curve + " 0 " + curve + " " + curve + " "
  // go down
  d += "l 0 " + (height) + " "
  //  curve down and left
  d += "q 0 " + curve + " -" + curve + " " + curve + " "
  // go left
  d += "l -" + (width / 2) + " 0 "
  // close
  d += " z"
  return d
}


//----------------------------------------------------------------------------//
//                                                                            //
//                                                                            //
//                                                                            //
//                                 MAKE PLOTS                                 //
//                                                                            //
//                                                                            //
//                                                                            //
//----------------------------------------------------------------------------//



//-------------------------------------------------------------------//
//                                                                   //
//                           FEATURE SELECTION                       //
//                                                                   //
//-------------------------------------------------------------------//
function make_fs_plot(models) {
  var svg, chart_group, data_extent, canvas, margins, max_draw_space, doc_state
  var plot = "feature_selection"
  var quantile_keys = ["0.00", "0.25", "0.50", "0.75", "1.00"]
  // extract current model
  var model = extractModel(models)
  // extract corresponding data
  var data = model[plot]
  data.sort(function(a, b) {return a.number_of_features - b.number_of_features})

  var output_container_id = charts_config.chart_container
  svg = make_chart_svg(plot)
  if (svg.select("g.plot_group").empty()) {
    chart_group = svg.append("g").attr("class", "plot_group")
    charts_config.document_state.current_plot = plot
  } else {
    chart_group = svg.select("g.plot_group")
  }
  chart_group.data(data)

  var data_extent, x_axis_label, y_axis_label
  data_extent = {
    "max": {
      "x": data.length + 1,
      "y": d3.max(data, function(d) {return d.quantiles["1.00"]})
    },
    "min": {
      "x": 0,
      "y": d3.min(data, function(d) {return d.quantiles["0.00"]})
    }
  }



  canvas = extract_canvas_from_svg(svg)
  margins = make_margins(chart_group, canvas, data_extent, true)
  max_draw_space = calculate_maximum_drawing_space(canvas, margins)
  make_title(chart_group, ["Feature Selection"], margins, canvas, max_draw_space)


  x_axis_label = "Features"
  y_axis_label = "Cross-validated accuracy"
  var custom_ticks = [""]
  data.map(function(d) { custom_ticks.push(d.number_of_features)})
  custom_ticks.push("")
  var x_scale = d3.scaleLinear().domain([data_extent.min.x, data_extent.max.x]).range([0, max_draw_space.x])
  var y_scale = d3.scaleLinear().domain([data_extent.max.y, data_extent.min.y]).range([0, max_draw_space.y])
  make_axes(chart_group, x_scale, y_scale, canvas, margins, max_draw_space, x_axis_label, y_axis_label, custom_ticks)

  // box parameters
  box = {}
  box.spacing = charts_config.plots.box_and_whiskers.spacing
  box.width = max_draw_space.x / (data.length + 1) - box.spacing

  var previous_boxes = chart_group.selectAll("g." + plot + "_box_group").size()
  var current_boxes = data.length

  var box_group = chart_group.selectAll("g." + plot + "_box_group").data(data)
  var whiskers = box_group.selectAll("g.whiskers")

  var not_opt_opac = 0.7

  if (previous_boxes < current_boxes) {
    var box_group = box_group.enter().append("g").attr("class", plot + "_box_group")
    box_group.selectAll("rect").data(function(d) {return [d,d,d,d]}).enter().append("rect")
             .attr("class", function(d,i) {
               if (i == 0 || i == 3) {
                 return plot + "_hidden_whisker_box_rect"
               } else {
                 return plot + "_quantile_box_rect"
               }})

    var whiskers = box_group.append("g").attr("class", "whiskers")
    whiskers.append("g").attr("class", "lower_whisker").append("path").attr("class", "whisker")
    whiskers.append("g").attr("class", "upper_whisker").append("path").attr("class", "whisker")
  } else if (previous_boxes > current_boxes) {
    box_group.exit().transition().delay(function(d, i) {return 100})
    .attr("transform","translate(0,0)").remove()
  } else {
    var box_group = chart_group.selectAll("g." + plot + "_box_group")
  }

  box_group = chart_group.selectAll("g." + plot + "_box_group")
  box_group.selectAll("rect").data(function(d) {return [d,d,d,d]})

    whiskers = box_group.selectAll("g.whiskers")
    box_group.each(function(dat, ind) {
      var current_quantile_group = d3.select(this)
      var quantile_data = {"0.00": dat.quantiles["0.00"], "0.25": dat.quantiles["0.25"], "0.50": dat.quantiles["0.50"],  "0.75": dat.quantiles["0.75"], "1.00": dat.quantiles["1.00"]}
      var low_box, high_box
       current_quantile_group.selectAll('rect').each(

         function(d, i){
           var current_rect = d3.select(this)
           var class_name = current_rect.node().className.baseVal

           var x_shift = box_x_shift(x_scale, ind, margins, box)
           var y_shift = box_y_shift(y_scale, quantile_data, i, margins, box)

           if (class_name.includes("_quantile_box")) {
             current_rect.attr("fill", charts_config.colors.palette[ind])
             .attr("stroke", charts_config.plots.box_and_whiskers.stroke)
             .style("opacity", function(d) {return d.optimal_number_of_features ? 1 : not_opt_opac})
             .transition().delay(function(d, i) {return 100 + ind * 10})
             .attr("height", box_height(y_scale, quantile_data, i))
             .attr("width", box.width)
             .attr("transform", "translate("+x_shift+","+y_shift+")")



           } else {
             current_rect.attr("fill", "white").style("opacity",0)
             .transition().delay(function(d, i) {return 100 + ind * 10})
             .attr("height", box_height(y_scale, quantile_data, i))
             .attr("width", box.width)
             .attr("transform", "translate("+x_shift+","+y_shift+")")

             if (i == 0) {
               low_box = {"x":x_shift,"y":y_shift,"h":box_height(y_scale, quantile_data, i),"w":box.width}
               lower_whisker = make_whisker_path(low_box, false)
               current_quantile_group.select("g.lower_whisker").select("path.whisker").transition().delay(function(d, i) {return 100 + ind * 10}).attr("d", lower_whisker).attr("fill", "none")
               .attr("stroke", charts_config.colors.palette[ind])
               .attr("stroke-width", charts_config.plots.box_and_whiskers.whiskers.width)
               .style("opacity", function(d) {return d.optimal_number_of_features ? 1 : not_opt_opac})
             } else if (i == 3) {
               high_box = {"x":x_shift,"y":y_shift,"h":box_height(y_scale, quantile_data, i),"w":box.width}
               upper_whisker = make_whisker_path(high_box, true)
               current_quantile_group.select("g.upper_whisker").select("path.whisker").transition().delay(function(d, i) {return 100 + ind * 10}).attr("d", upper_whisker).attr("fill", "none")
               .attr("stroke", charts_config.colors.palette[ind])
               .attr("stroke-width", charts_config.plots.box_and_whiskers.whiskers.width)
               .style("opacity", function(d) {return d.optimal_number_of_features ? 1 : not_opt_opac})
             }
           }
         }
       )
    })


    // box_group.each(function(dat, ind) {
    //   var current_quantile_group = d3.select(this)
    //   current_quantile_group.selectAll('rect').each(function(d, i) {
    //     var current_rect = d3.select(this)
    //     var class_name = current_rect.node().className.baseVal
    //     if (class_name.includes("_quantile_box")) {
    //       current_rect.raise()
    //     }
    //   })
    // })

    box_group.selectAll("rect").on("mouseover", mouseoverFunction)
    box_group.selectAll("rect").on("mouseout", mouseoutFunction)
    function mouseoverFunction(d, i) {

      d3.select(this).style("opacity", function () {if (i != 0 && i != 3) {return 1} else {return 0}})
      if (i == 0) {
        d3.select(d3.select(this).node().parentNode)
        .select("g.lower_whisker").select("path.whisker")
        .style("opacity", 1)
      } else if (i == 3) {
        d3.select(d3.select(this).node().parentNode)
        .select("g.upper_whisker").select("path.whisker")
        .style("opacity", 1)
      }


      var quantile_data = {"0.00": d.quantiles["0.00"], "0.25": d.quantiles["0.25"], "0.50": d.quantiles["0.50"],  "0.75": d.quantiles["0.75"], "1.00": d.quantiles["1.00"]}
      var box = d3.select(this).node().getBBox()
      y = box.y + box.height / 2
      x = x_scale(data.indexOf(d)) + margins.x.left + margins.axes.y + box.width / 2

      var tooltip_text = [
        "Features: " + d.number_of_features,
        "Q4: " + quantile_data["1.00"].toFixed(3),
        "Q3: " + quantile_data["0.75"].toFixed(3),
        "Q2: " + quantile_data["0.50"].toFixed(3),
        "Q1: " + quantile_data["0.25"].toFixed(3),
        "Q0: " + quantile_data["0.00"].toFixed(3)
      ]

      make_tooltip(d3.select(this), tooltip_text, chart_group, canvas, margins)
    }
    function mouseoutFunction(d, i) {
        chart_group.select("g.tooltip_group").remove()
        d3.select(this).style("opacity", function () {
          if (i != 0 && i != 3) {
            if (d.optimal_number_of_features) {
              return 1
            } else {
              return not_opt_opac
            }
          } else {
            return 0
          }
        })

        if (i == 0) {
          d3.select(d3.select(this).node().parentNode)
          .select("g.lower_whisker").select("path.whisker")
          .style("opacity", not_opt_opac)
        } else if (i == 3) {
          d3.select(d3.select(this).node().parentNode)
          .select("g.upper_whisker").select("path.whisker")
          .style("opacity", not_opt_opac)
        }
  }

}




function box_height(y_scale, extracted_data, i) {
  var data_keys = keys(extracted_data)
  return y_scale(extracted_data[data_keys[i]]) - y_scale(extracted_data[data_keys[i+1]])
}

function box_x_shift(x_scale, i, margins, box) {
  var x_shift = x_scale(i) + margins.x.left + margins.axes.y + box.width / 2 + box.spacing
  return x_shift
}

function box_y_shift(y_scale, extracted_data, i, margins, box) {
  var data_keys = keys(extracted_data)
  var sum = 0
  for (var j = 3; j > i; j--) {
    sum += y_scale(extracted_data[data_keys[j]]) - y_scale(extracted_data[data_keys[j+1]])
  }
  var y_shift = margins.y.top + margins.title + margins.buttons + sum + y_scale(extracted_data[data_keys[4]])
  return y_shift
}

function make_whisker_path(box, which) {
  var x, y, w, h
  x = box.x
  y = box.y
  h = box.h
  w = box.w
 var p
 if (which) { // make upper whisker
   p = "M " + (x + w /2) + " " + (y + h) + " "
   p += "L " + (x + w /2) + " " + (y) + " "
   p += "L " + (x) + " " + (y) + " "
   p += "L " + (x + w) + " " + (y) + " "
 } else { // make lower whisker
   p = "M " + (x + w /2) + " " + (y) + " "
   p += "L " + (x + w /2) + " " + (y + h) + " "
   p += "L " + (x) + " " + (y + h) + " "
   p += "L " + (x + w) + " " + (y + h) + " "
 }
 return p
}

//-------------------------------------------------------------------//
//                                                                   //
//                          OUT OF BAG ERROR                         //
//                                                                   //
//-------------------------------------------------------------------//
function make_oob_plot(models) {
  var svg, chart_group, data_extent, canvas, margins, max_draw_space, doc_state
  var plot = "out_of_bag_error"
  // extract current model
  var model = extractModel(models)
  // extract corresponding data
  var data = model[plot]
  // Make SVG element and group for the chart
  var output_container_id = charts_config.chart_container

  svg = make_chart_svg(plot)
  if (svg.select("g.plot_group").empty()) {
    chart_group = svg.append("g").attr("class", "plot_group")
    charts_config.document_state.current_plot = plot
  } else {
    chart_group = svg.select("g.plot_group")
  }

  var data_extent, x_axis_label, y_axis_label
  data_extent = {
    "max": {
      "x": d3.max(data, function(d) {return d.trees}),
      "y": d3.max(data, function(d) {return Math.max(d.conditions.treatment, d.conditions.control, d.conditions.both)})
    },
    "min": {
      "x": d3.min(data, function(d) {return d.trees}),
      "y": d3.min(data, function(d) {return Math.min(d.conditions.treatment, d.conditions.control, d.conditions.both)})
    }
  }

  x_axis_label = "Trees"
  y_axis_label = "Error"

  canvas = extract_canvas_from_svg(svg)
  margins = make_margins(chart_group, canvas, data_extent, true)
  max_draw_space = calculate_maximum_drawing_space(canvas, margins)
  make_title(chart_group, ["OOB error rate vs number of trees"], margins, canvas, max_draw_space)
  var x_scale = d3.scaleLinear().domain([data_extent.min.x, data_extent.max.x]).range([0, max_draw_space.x])
  var y_scale = d3.scaleLinear().domain([data_extent.max.y, data_extent.min.y]).range([0, max_draw_space.y])
  make_axes(chart_group, x_scale, y_scale, canvas, margins, max_draw_space, x_axis_label, y_axis_label)


  var lineKeys = d3.keys(data[0].conditions)
  // If first call, add the line
  if (chart_group.selectAll("g." + plot + "_line_group").empty()) {
    // Add the group for the line
    chart_group.append("g").attr("class", plot + "_line_group")
    chart_group.select("g." + plot + "_line_group").selectAll("g.line")
              .data(lineKeys)
              .enter()
              .append("g")
              .attr("class", "line")
              .append("path")
              .attr("class", function(d) {return "line_" + d})

    chart_group.append("g").attr("class", plot + "_point_group")
    chart_group.select("g." + plot + "_point_group").selectAll("g.points")
              .data(lineKeys)
              .enter()
              .append("g")
              .attr("class", "hidden_points")
              .selectAll("circle")
              .data(function(d, i) {
                return values(data).map(
                  function(dat, ind) {
                    // console.log("setting point data", dat.trees);
                    return {"x": dat.trees, "y": dat.conditions[d], "condition": d}
                  })
                })
              .enter()
              .append("circle")
              .attr("class", function(d) {
                // console.log("testing data setting", d)
                return "circle_" + d.condition
              })


  }

  //


  // UPDATE DATA

  chart_group.select("g." + plot + "_point_group")
             .selectAll("g.hidden_points")
             .each(function (d,i) {
               var current_point_group = d3.select(this).selectAll("circle")
               current_point_group.each(function(dat, ind) {
                 var current_data = d3.select(this).datum()
                 current_data.y = data[ind].conditions[d]
               })
             })



  // Double mapped function, first to groups contain each a path svg element
  // and then mapped amongst those paths to construct the lines.
  chart_group.select("g." + plot + "_line_group").selectAll("g.line")
            .each(function(d,i){
              var currentLineGroup = d3.select(this).select("path").data([data])
              var curCond = d
              var curLine = i
              var line = d3.line()
                           .x(function (d) {return x_scale(d.trees) + margins.x.left + margins.axes.y})
                           .y(function (d) {return y_scale(d.conditions[curCond])+ margins.y.top + margins.title + margins.buttons})
              currentLineGroup.attr("d", line)
                              .attr("fill", "none")
                              .attr("stroke", function(d,i) {return charts_config.colors.palette[curLine * 3]})
                              .attr("stroke-width", charts_config.plots.line.width)

            })


  // Double mapped function, first to groups contain each a set of points (circle svg elements)
  // and then mapped amongst those circles.
  chart_group.select("g." + plot + "_point_group").selectAll("g.hidden_points")
            .each(function (d,i) {
              var currentCircleGroup = d3.select(this).selectAll("circle")
              currentCircleGroup.attr("r", charts_config.plots.line["hidden_radius"])
              .attr("fill", function(dat, ind) {return charts_config.colors.palette[i * 3]})
              .style("opacity", 0)
              .on("mouseover", mouseoverFunction)
              .on("mouseout", mouseoutFunction)
              .attr("transform", function(d, i) {return "translate(" + (x_scale(d.x) + margins.x.left + margins.axes.y) + "," + (y_scale(d.y)+ margins.y.top + margins.title + margins.buttons) + ")"})
              // .raise()
              // .attr("cx", function (dat, ind) {return (x_scale(dat.x) + margins.x.left + margins.axes.y)})
              // .attr("cy", function (dat, ind) {console.log(dat.y, y_scale(dat.y)); return (y_scale(dat.y)+ margins.y.top + margins.title + margins.buttons)})
            })


  function mouseoverFunction(d, i) {
      d3.select(this).style("opacity", charts_config.plots.line.opacity.hover)
      x = x_scale(d.x) + margins.x.left + margins.axes.y
      y = y_scale(d.y) + margins.y.top + margins.title + margins.buttons
      make_tooltip(d3.select(this), [d.x,  d.condition + ": "+ d.y], chart_group, canvas, margins)
  }
  function mouseoutFunction(d, i) {
      chart_group.select("g.tooltip_group").remove()
      d3.select(this).style("opacity", 0)
  }

}

//-------------------------------------------------------------------//
//                                                                   //
//                             OUTLIERS                              //
//                                                                   //
//-------------------------------------------------------------------//
function make_outlier_plot(models) {
  var svg, chart_group, data_extent, canvas, margins, max_draw_space, doc_state
  var plot = "outliers"
  var button_keys = ["outlier", "cross_validation_error"]
  // extract current model
  var model = extractModel(models)
  // extract corresponding data
  var data = model[plot]

  svg = make_chart_svg(plot)
  if (svg.select("g.plot_group").empty()) {
    chart_group = svg.append("g").attr("class", "plot_group")
    charts_config.document_state.current_plot = plot
    charts_config.document_state.outlier_plot = button_keys[0]
  } else {
    chart_group = svg.select("g.plot_group")
  }

  doc_state = charts_config.document_state.outlier_plot
  var data_extent, x_axis_label, y_axis_label
  data_extent = {
    "max": {
      "x": data.length,
      "y": d3.max(data, function(d) {return d[doc_state]})
    },
    "min": {
      "x": 1,
      "y": d3.min(data, function(d) {return d[doc_state]})
    }
  }

  // console.log(data)


  x_axis_label = "Samples"
  if (doc_state == "outlier") {
    y_axis_label = "Outlier Score"
    title_label = "Outlier Plot"
  } else {
    y_axis_label = "Cross-validation Error"
    title_label = "Cross-validation Error Plot"
  }

  canvas = extract_canvas_from_svg(svg)
  margins = make_margins(chart_group, canvas, data_extent, true)
  max_draw_space = calculate_maximum_drawing_space(canvas, margins)
  make_buttons(chart_group, margins, canvas, max_draw_space, button_keys, outlier_button_click)
  color_buttons(chart_group, doc_state)
  make_title(chart_group, [title_label], margins, canvas, max_draw_space)
  var x_scale = d3.scaleLinear().domain([data_extent.min.x, data_extent.max.x]).range([0, max_draw_space.x])
  var y_scale = d3.scaleLinear().domain([data_extent.max.y, data_extent.min.y]).range([0, max_draw_space.y])
  make_axes(chart_group, x_scale, y_scale, canvas, margins, max_draw_space, x_axis_label, y_axis_label)

  var previous_points = chart_group.selectAll("g." + plot + "_scatter_group").size()
  var current_points = data.length
  var scatter_group = chart_group.selectAll("g." + plot + "_scatter_group").data(data)


  var groups = extract_all_groups(data, "group")

  if (previous_points < current_points) {
    // Add the points
    chart_group.selectAll("g." + plot + "_scatter_group")
              .data(data)
              .enter()
              .append("g")
              .attr("class", plot + "_scatter_group")
              .append("circle")
              .attr("r", charts_config.plots.scatter.point.radius)
              .attr("fill", function(d, i) {return charts_config.colors.palette[groups.indexOf(d.group) * 5]})
              .style("opacity", charts_config.plots.scatter.opacity.non_hover)
  } else if (previous_points > current_points) {
    scatter_group.exit().transition().delay(function(d, i) {return 100})
    .attr("transform", "translate(0,0)").remove()
  } else {

  }

  scatter_group = d3.selectAll("g." + plot + "_scatter_group")
  scatter_group.transition().delay(function(d, i) {return 100})
               .attr("transform", function(d, i) {return "translate(" + (x_scale(i + 1) + margins.x.left + margins.axes.y) + "," + (y_scale(d[doc_state])+ margins.y.top + margins.title + margins.buttons) + ")"})


 scatter_group.on("mouseover", mouseoverFunction)
 scatter_group.on("mouseout", mouseoutFunction)

 function mouseoverFunction(d, i) {
  var d = d3.select(this).datum()
  make_tooltip(d3.select(this), ["Sample: " + d.sample, "Group: " + d.group, "Score: " + d[doc_state]], chart_group, canvas, margins)
 }

 function mouseoutFunction(d, i) {
  chart_group.select("g.tooltip_group").remove()
 }

}



function outlier_button_click() {
  var current = d3.select(this).datum()
  charts_config.document_state.outlier_plot = current
  make_outlier_plot(models)
}

//-------------------------------------------------------------------//
//                                                                   //
//                     PRINCIPAL COMPONENT ANALYSIS                  //
//                                                                   //
//-------------------------------------------------------------------//
function make_pca_plot(models) {
  var svg, chart_group, data_extent, canvas, margins, max_draw_space, doc_state
  var plot = "principal_component_analysis"
  // extract current model
  var model = extractModel(models)
  // extract corresponding data
  var data = model[plot].samples

  var output_container_id = charts_config.chart_container

  svg = make_chart_svg(plot)
  if (svg.select("g.plot_group").empty()) {
    chart_group = svg.append("g").attr("class", "plot_group")
    charts_config.document_state.current_plot = plot
  } else {
    chart_group = svg.select("g.plot_group")
  }

  var data_extent, x_axis_label, y_axis_label
  data_extent = {
    "max": {
      "x": d3.max(data, function(d) {return d.distance.x}),
      "y": d3.max(data, function(d) {return d.distance.y})
    },
    "min": {
      "x": d3.min(data, function(d) {return d.distance.x}),
      "y": d3.min(data, function(d) {return d.distance.y})
    }
  }

  x_axis_label = "PC 1: " + model[plot]["component_variance"]["principal_component_1"] + "%"
  y_axis_label = "PC 2: " + model[plot]["component_variance"]["principal_component_2"] + "%"
  canvas = extract_canvas_from_svg(svg)
  margins = make_margins(chart_group, canvas, data_extent, true)
  max_draw_space = calculate_maximum_drawing_space(canvas, margins)
  make_title(chart_group, ["PCA Plot"], margins, canvas, max_draw_space)
  var x_scale = d3.scaleLinear().domain([data_extent.min.x, data_extent.max.x]).range([0, max_draw_space.x])
  var y_scale = d3.scaleLinear().domain([data_extent.max.y, data_extent.min.y]).range([0, max_draw_space.y])
  make_axes(chart_group, x_scale, y_scale, canvas, margins, max_draw_space, x_axis_label, y_axis_label)

  var previous_points = chart_group.selectAll("g." + plot + "_scatter_group").size()
  var current_points = data.length
  var scatter_group = chart_group.selectAll("g." + plot + "_scatter_group").data(data)


  var groups = extract_all_groups(data, "group")

  if (previous_points < current_points) {
    // Add the points
    chart_group.selectAll("g." + plot + "_scatter_group")
              .data(data)
              .enter()
              .append("g")
              .attr("class", plot + "_scatter_group")
              .append("circle")
              .attr("r", charts_config.plots.scatter.point.radius)
              .attr("fill", function(d, i) {return charts_config.colors.palette[groups.indexOf(d.group) * 5]})
              .style("opacity", charts_config.plots.scatter.opacity.non_hover)
  } else if (previous_points > current_points) {
    scatter_group.exit().transition().delay(function(d, i) {return 100})
    .attr("transform", "translate(0,0)").remove()
  } else {

  }

  scatter_group = d3.selectAll("g." + plot + "_scatter_group")
  scatter_group.transition().delay(function(d, i) {return 100})
               .attr("transform", function(d, i) {return "translate(" + (x_scale(d.distance.x) + margins.x.left + margins.axes.y) + "," + (y_scale(d.distance.y)+ margins.y.top + margins.title + margins.buttons) + ")"})


               scatter_group.on("mouseover", mouseoverFunction)
               scatter_group.on("mouseout", mouseoutFunction)

 function mouseoverFunction(d, i) {
  var d = d3.select(this).datum()
  make_tooltip(d3.select(this), ["Sample: " + d.sample, "Group: " + d.group], chart_group, canvas, margins)
 }

 function mouseoutFunction(d, i) {
  chart_group.select("g.tooltip_group").remove()
 }

}

//-------------------------------------------------------------------//
//                                                                   //
//                           PRECISION RECALL                        //
//                                                                   //
//-------------------------------------------------------------------//
function make_pr_plot(models) {
  var svg, chart_group, data_extent, canvas, margins, max_draw_space, doc_state
  var plot = "precision_recall"
  var plot_cv = "precision_recall_cross_validation_error"
  // extract current model
  var model = extractModel(models)
  // extract corresponding data
  var data = model[plot]
  var data_cv = model[plot_cv]

  var output_container_id = charts_config.chart_container
  svg = make_chart_svg(plot)
  if (svg.select("g.plot_group").empty()) {
    chart_group = svg.append("g").attr("class", "plot_group")
    charts_config.document_state.current_plot = plot
  } else {
    chart_group = svg.select("g.plot_group")
  }

  var data_extent, data_extent_cv, x_axis_label, y_axis_label, comb_data_ext
  data_extent = {
    "max": {
      "x": d3.max(data, function(d) {return d.x}),
      "y": d3.max(data, function(d) {return d.y})
    },
    "min": {
      "x": d3.min(data, function(d) {return d.x}),
      "y": d3.min(data, function(d) {return d.y})
    }
  }


  data_extent_cv = {
    "max": {
      "x": d3.max(data_cv, function(d) {return d.x}),
      "y": d3.max(data_cv, function(d) {return d.y})
    },
    "min": {
      "x": d3.min(data_cv, function(d) {return d.x}),
      "y": d3.min(data_cv, function(d) {return d.y})
    }
  }

  comb_data_ext = {
    "max": {
      "x": Math.max(data_extent.max.x, data_extent_cv.max.x),
      "y": Math.max(data_extent.max.y, data_extent_cv.max.y)
    },
    "min": {
      "x": Math.min(data_extent.min.x, data_extent_cv.min.x),
      "y": Math.min(data_extent.min.y, data_extent_cv.min.y)
    }
  }


  x_axis_label = "Recall"
  y_axis_label = "Precision"
  canvas = extract_canvas_from_svg(svg)
  margins = make_margins(chart_group, canvas, comb_data_ext, true)
  max_draw_space = calculate_maximum_drawing_space(canvas, margins)
  make_title(chart_group, ["Precision Recall Plot"], margins, canvas, max_draw_space)
  var x_scale = d3.scaleLinear().domain([comb_data_ext.min.x, comb_data_ext.max.x]).range([0, max_draw_space.x])
  var y_scale = d3.scaleLinear().domain([comb_data_ext.max.y, comb_data_ext.min.y]).range([0, max_draw_space.y])
  make_axes(chart_group, x_scale, y_scale, canvas, margins, max_draw_space, x_axis_label, y_axis_label)

  var line = d3.line().x(function (d) {return x_scale(d.x) + margins.x.left + margins.axes.y}).y(function (d) {return y_scale(d.y)+ margins.y.top + margins.title + margins.buttons})
  var line_cv = d3.line().x(function (d) {return x_scale(d.x) + margins.x.left + margins.axes.y}).y(function (d) {return y_scale(d.y)+ margins.y.top + margins.title + margins.buttons})
  if (chart_group.selectAll("g." + plot + "_line_group").empty()) {
    // Add the group for the line
    chart_group.append("g").attr("class", plot + "_line_group").append("path")

    chart_group.selectAll("g." + plot + "_point_group")
              .data(data)
              .enter()
              .append("g")
              .attr("class", plot + "_point_group")
              .append("circle")
              .attr("r", charts_config.plots.line.hidden_radius)
              .style("opacity", 0)


    // Add the group for the line_cv
    chart_group.append("g").attr("class", plot_cv + "_line_group").append("path")

    chart_group.selectAll("g." + plot_cv + "_point_group")
              .data(data_cv)
              .enter()
              .append("g")
              .attr("class", plot_cv + "_point_group")
              .append("circle")
              .attr("r", charts_config.plots.line.hidden_radius)
              .style("opacity", 0)

  }

  var line_group = d3.selectAll("g." + plot + "_line_group").data(data)
  var line_cv_group = d3.selectAll("g." + plot_cv + "_line_group").data(data_cv)
  line_group.select("path")
            .data([data])
            .attr("class", "line")
            .attr("d", line)
            .attr("fill", charts_config.plots.line.fill)
            .attr("stroke", charts_config.plots.line.stroke)
            .attr("stroke-width", charts_config.plots.line.width)

  line_cv_group.select("path")
            .data([data_cv])
            .attr("class", "line")
            .attr("d", line_cv)
            .attr("fill", charts_config.plots.line.fill)
            .attr("stroke", charts_config.colors.palette[0])
            .attr("stroke-width", charts_config.plots.line.width)

  var point_group = d3.selectAll("g." + plot + "_point_group").data(data)
  var point_cv_group = d3.selectAll("g." + plot_cv + "_point_group").data(data_cv)

  point_group.attr("transform", function(d, i) {return "translate(" + (x_scale(d.x) + margins.x.left + margins.axes.y) + "," + (y_scale(d.y)+ margins.y.top + margins.title + margins.buttons) + ")"})
  point_cv_group.attr("transform", function(d, i) {return "translate(" + (x_scale(d.x) + margins.x.left + margins.axes.y) + "," + (y_scale(d.y)+ margins.y.top + margins.title + margins.buttons) + ")"})

  point_group.on("mouseover", mouseoverFunction)
  point_group.on("mouseout", mouseoutFunction)
  point_cv_group.on("mouseover", mouseoverFunction)
  point_cv_group.on("mouseout", mouseoutFunction)

 function mouseoverFunction(d, i) {
   var class_name = d3.select(this).node().className.baseVal
   var this_circle = d3.select(this).select("circle")
   this_circle.style("opacity", charts_config.plots.line.opacity.hover).style("fill", charts_config.plots.line.stroke)
   var tooltip_text = "PR"
   if (class_name.includes("cross")) {
     this_circle.style("fill", charts_config.colors.palette[0])
     tooltip_text = "PR_CV"
   }
   x = x_scale(d.x) + margins.x.left + margins.axes.y
   y = y_scale(d.y) + margins.y.top + margins.title + margins.buttons
   make_tooltip(d3.select(this), [tooltip_text, "Recall: " + d.x, "Precision: " + d.y], chart_group, canvas, margins)
 }

 function mouseoutFunction(d, i) {
  chart_group.select("g.tooltip_group").remove()
  d3.select(this).select("circle").style("opacity", 0)
 }

}
//-------------------------------------------------------------------//
//                                                                   //
//                 RECEIVER OPERATING CHARACTERISTICS                //
//                                                                   //
//-------------------------------------------------------------------//
function make_roc_plot(models) {
  var svg, chart_group, data_extent, canvas, margins, max_draw_space, doc_state
  var plot = "receiver_operating_characteristic"
  // extract current model
  var model = extractModel(models)
  // extract corresponding data
  var data = model[plot]["true_positive_rate"]

  var output_container_id = charts_config.chart_container
  svg = make_chart_svg(plot)
  if (svg.select("g.plot_group").empty()) {
    chart_group = svg.append("g").attr("class", "plot_group")
    charts_config.document_state.current_plot = plot
  } else {
    chart_group = svg.select("g.plot_group")
  }

  var data_extent, x_axis_label, y_axis_label
  data_extent = {
    "max": {
      "x": d3.max(data, function(d) {return d.x}),
      "y": d3.max(data, function(d) {return d.y})
    },
    "min": {
      "x": d3.min(data, function(d) {return d.x}),
      "y": d3.min(data, function(d) {return d.y})
    }
  }

  x_axis_label = "False positive rate"
  y_axis_label = "True positive rate"
  canvas = extract_canvas_from_svg(svg)
  margins = make_margins(chart_group, canvas, data_extent, true)
  max_draw_space = calculate_maximum_drawing_space(canvas, margins)
  make_title(chart_group, ["ROC Plot"], margins, canvas, max_draw_space)
  var x_scale = d3.scaleLinear().domain([data_extent.min.x, data_extent.max.x]).range([0, max_draw_space.x])
  var y_scale = d3.scaleLinear().domain([data_extent.max.y, data_extent.min.y]).range([0, max_draw_space.y])
  make_axes(chart_group, x_scale, y_scale, canvas, margins, max_draw_space, x_axis_label, y_axis_label)

  var line = d3.line().x(function (d) {return x_scale(d.x) + margins.x.left + margins.axes.y}).y(function (d) {return y_scale(d.y)+ margins.y.top + margins.title + margins.buttons})
  if (chart_group.selectAll("g." + plot + "_line_group").empty()) {
    // Add the group for the line
    chart_group.append("g").attr("class", plot + "_line_group").append("path")

    chart_group.selectAll("g." + plot + "_point_group")
              .data(data)
              .enter()
              .append("g")
              .attr("class", plot + "_point_group")
              .append("circle")
              .attr("r", charts_config.plots.line.hidden_radius)
              .style("opacity", 0)

  }

  var line_group = d3.selectAll("g." + plot + "_line_group").data(data)
  line_group.select("path")
            .data([data])
            .attr("class", "line")
            .attr("d", line)
            .attr("fill", charts_config.plots.line.fill)
            .attr("stroke", charts_config.plots.line.stroke)
            .attr("stroke-width", charts_config.plots.line.width)

  var point_group = d3.selectAll("g." + plot + "_point_group").data(data)
  point_group.attr("transform", function(d, i) {return "translate(" + (x_scale(d.x) + margins.x.left + margins.axes.y) + "," + (y_scale(d.y)+ margins.y.top + margins.title + margins.buttons) + ")"})

  point_group.on("mouseover", mouseoverFunction)
  point_group.on("mouseout", mouseoutFunction)

 function mouseoverFunction(d, i) {
   d3.select(this).select("circle").style("opacity", charts_config.plots.line.opacity.hover).style("fill", charts_config.plots.line.stroke)
   x = x_scale(d.x) + margins.x.left + margins.axes.y
   y = y_scale(d.y) + margins.y.top + margins.title + margins.buttons
   make_tooltip(d3.select(this), ["FPR: " + d.x, "TPR: "+ d.y], chart_group, canvas, margins)
 }

 function mouseoutFunction(d, i) {
  chart_group.select("g.tooltip_group").remove()
  d3.select(this).select("circle").style("opacity", 0)
 }

}

//-------------------------------------------------------------------//
//                                                                   //
//                         VARIABLE IMPORTANCE                       //
//                                                                   //
//-------------------------------------------------------------------//
function make_vi_plot(models) {
  var svg, chart_group, data_extent, canvas, margins, max_draw_space, doc_state
  var plot = "variable_importance"
  // extract current model
  var model = extractModel(models)
  // extract corresponding data
  var data = model[plot]
  // Sort data
  data.sort(function(a, b) {return a["gini_score"] - b["gini_score"]})
  // Make SVG element and group for the chart
  var output_container_id = charts_config.chart_container

  svg = make_chart_svg(plot)
  if (svg.select("g.plot_group").empty()) {
    chart_group = svg.append("g").attr("class", "plot_group")
    charts_config.document_state.current_plot = plot
  } else {
    chart_group = svg.select("g.plot_group")
  }

  var data_extent, x_axis_label, y_axis_label
  data_extent = {
    "max": {
      "x": d3.max(data, function(d) {return d["gini_score"]}),
      "y": data.length
    },
    "min": {
      "x": d3.min(data, function(d) {return d["gini_score"]}),
      "y": 0
    }
  }

  x_axis_label = "Mean Decrease Gini"
  y_axis_label = "Variables"

  canvas = extract_canvas_from_svg(svg)
  margins = make_margins(chart_group, canvas, data_extent, true)
  max_draw_space = calculate_maximum_drawing_space(canvas, margins)
  make_title(chart_group, ["Variable Importance"], margins, canvas, max_draw_space)
  var x_scale = d3.scaleLinear().domain([data_extent.min.x, data_extent.max.x]).range([0, max_draw_space.x])
  var y_scale = d3.scaleLinear().domain([data_extent.max.y, data_extent.min.y]).range([0, max_draw_space.y])
  make_axes(chart_group, x_scale, y_scale, canvas, margins, max_draw_space, x_axis_label, y_axis_label)

  var previous_points = chart_group.selectAll("g." + plot + "_scatter_group").size()
  var current_points = data.length
  var scatter_group = chart_group.selectAll("g." + plot + "_scatter_group").data(data)


  var groups = extract_all_groups(data, "variable")

  if (previous_points < current_points) {
    // Add the points
    chart_group.selectAll("g." + plot + "_scatter_group")
              .data(data)
              .enter()
              .append("g")
              .attr("class", plot + "_scatter_group")
              .append("circle")
              .attr("r", charts_config.plots.scatter.point.radius)
              .attr("fill", function(d, i) {var ind = groups.indexOf(d.variable); if (ind > charts_config.colors.palette.length) {ind = ind % charts_config.colors.palette.length;}; return charts_config.colors.palette[ind]})
              .style("opacity", charts_config.plots.scatter.opacity.non_hover)
  } else if (previous_points > current_points) {
    scatter_group.exit().transition().delay(function(d, i) {return 100})
    .attr("transform", "translate(0,0)")
    scatter_group.exit().remove()
  } else {

  }

  scatter_group = d3.selectAll("g." + plot + "_scatter_group")
  scatter_group.transition().delay(function(d, i) {return 100})
               .attr("transform", function(d, i) {return "translate(" + (x_scale(d.gini_score) + margins.x.left + margins.axes.y) + "," + (y_scale(i + 1)+ margins.y.top + margins.title + margins.buttons) + ")"})


 scatter_group.on("mouseover", mouseoverFunction)
 scatter_group.on("mouseout", mouseoutFunction)

 function mouseoverFunction(d, i) {
  var d = d3.select(this).datum()
  make_tooltip(d3.select(this), [d.variable, d.gini_score], chart_group, canvas, margins)
 }

 function mouseoutFunction(d, i) {
  chart_group.select("g.tooltip_group").remove()
 }

}

//----------------------------------------------------------------------------//
//                                                                            //
//                                                                            //
//                                                                            //
//                      PATCHWORK SOLUTION FOR OLD JSON                       //
//                                                                            //
//                                                                            //
//                                                                            //
//----------------------------------------------------------------------------//
//-------------------------------------------------------------------//
//                                                                   //
//                       TRANSFORM OLD JSON                          //
//                                                                   //
//-------------------------------------------------------------------//
function transformOldJSON(json) {
  var pca_total, pca_variance, outliers, error_cv, error, variables, roc, pr, prcv, features, features_new
  pca_total = json.pca_total
  pca_variance = json.pca_variance
  outliers = json.outliers
  error_cv = json.error_cv_samples
  error = json.error
  variables = json.variables
  roc = json.roc
  pr = json.pr
  prcv = json.pr_cv
  features = json.features
  features_new = json.features_new

  var data = {
    "statistics": {},
    "principal_component_analysis": {"samples": [], "component_variance": {}},
    "outliers": [],
    "receiver_operating_characteristic": {"true_positive_rate": []},
    "precision_recall": [],
    "precision_recall_cross_validation_error": [],
    "variable_importance": [],
    "feature_selection": [],
    "feature_selection_new": [],
    "out_of_bag_error": []
  }

  // Add statistics
  data.statistics = json.statistics

  // Add component values to PCA object
  data.principal_component_analysis.component_variance.prinicpal_component_1 = pca_variance[0].variance
  data.principal_component_analysis.component_variance.prinicpal_component_2 = pca_variance[1].variance

  // Add distance sample values to PCA object
  for (var i = 0; i < pca_total.length; i++) {
    var obj = pca_total[i]
    var sample = {}
    sample.sample = obj.key
    sample.group = obj.values.group
    sample.distance = {}
    sample.distance.x = obj.values.x
    sample.distance.y = obj.values.y
    data.principal_component_analysis.samples.push(sample)
  }

  // Add outlier values to PCA object

  // Sort outlier and cross-val. error by sample so they are in same order
  error_cv.sort(function(a, b) {return a.label - b.label})
  outliers.sort(function(a, b) {return a.key - b.key})

  for (var i = 0; i < outliers.length; i++) {
    var obj = outliers[i]
    var sample = {}
    sample.outlier = obj.values.y
    sample.sample = obj.key
    sample.group = obj.values.group
    var obj = error_cv[i]
    sample.cross_validation_error = obj.value
    data.outliers.push(sample)
  }

  // Add OOB values
  var t = error[0].values
  var c =error[1].values
  var b =error[2].values
  for (var i = 0; i < t.length; i++) {
    var tree = {}

    tree.trees = t[i].x
    tree.conditions = {
      "treatment": t[i].y,
      "control": c[i].y,
      "both": b[i].y
    }

    data.out_of_bag_error.push(tree)
  }

  // Add variables values
  for (var i = 0; i < variables.length; i++) {
    var obj = variables[i]
    var importance = {}
    importance.variable = obj.key
    importance.gini_score = obj.values.x
    if (importance.gini_score !="NA") {
      data.variable_importance.push(importance)
    }
  }

  // Add roc values
  for (var i = 0; i < roc.length; i++) {
    var obj = roc[i]
    var point = {}
    point.x = obj.values.x
    point.y = obj.values.y
    data.receiver_operating_characteristic.true_positive_rate.push(point)
  }

  // Add PR values
  for (var i = 0; i < pr.length; i++) {
    var obj = pr[i]
    var point = {}

    point.x = obj.values.x
    point.y = obj.values.y

    data.precision_recall.push(point)
  }


  // Add PR_CV values
  for (var i = 0; i < prcv.length; i++) {
    var obj = prcv[i]
    var point = {}

    point.x = obj.values.x
    point.y = obj.values.y

    data.precision_recall_cross_validation_error.push(point)
  }

  // Add feature values
  for (var i = 0; i < features.length; i++) {
    var obj = features[i]
    var feature = {}
    feature.quantiles = {}

    feature.number_of_features = obj.label
    feature.optimal_number_of_features = obj.optimal
    feature.quantiles["0.25"] = obj.values["Q1"]
    feature.quantiles["0.50"] = obj.values["Q2"]
    feature.quantiles["0.75"] = obj.values["Q3"]
    feature.min = obj.values.whisker_low
    feature.max = obj.values.whisker_high

    data.feature_selection.push(feature)
  }

  // // Add new feature values
  // for (var i = 0; i < features_new.length; i++) {
  //   var obj = features_new[i]
  //   var feature_new = {}
  //   feature_new.quantiles = {}
  //
  //   feature_new.number_of_features = obj.label
  //   feature_new.optimal_number_of_features = obj.optimal
  //   feature_new.quantiles["0.00"] = obj.values.whisker_low
  //   feature_new.quantiles["0.25"] = obj.values["Q1"]
  //   feature_new.quantiles["0.50"] = obj.values["Q2"]
  //   feature_new.quantiles["0.75"] = obj.values["Q3"]
  //   feature_new.quantiles["1.00"] = obj.values.whisker_high
  //
  //   data.feature_selection_new.push(feature_new)
  // }


  return data
}




//----------------------------------------------------------------------------//
//                                                                            //
//                                                                            //
//                                                                            //
//                        CLASSIFICATION RESULTS TABLE                        //
//                                                                            //
//                                                                            //
//                                                                            //
//----------------------------------------------------------------------------//
//
// function classification_results_table(){
//   var chart_id = "oasis_classification_feature_importance_table"
//   var model_folder = "ClassificationResults/RandomForest/"
//   if (charts_config.document_state.balanced) {
//     model_folder += "balanced/"
//   } else {
//     model_folder += "unbalanced/"
//   }
//
//   if (charts_config.document_state.optimized) {
//     model_folder += "optimized_variables/"
//   } else {
//     model_folder += "all_variables/"
//   }
//
//   model_folder += "classification_table.tsv"
//
//
//
//
//
//   d3.tsv(model_folder, function(error, data) {
//     data.sort(function(a, b) {
//       return parseFloat(a["Feature Importance"]) - parseFloat(b["Feature Importance"])
//     })
//     data.reverse()
//
//     var table = d3.select("#" + chart_id)
//     table.selectAll("*").remove()
//     var table_head = table.append("thead")
//     var table_body = table.append("tbody")
//
//     var data_keys = keys(data[0])
//     var row, col_el
//
//     row = table_head.append("tr")
//
//     for (var i = 0; i < data_keys.length; i++) {
//       col_el = row.append("th")
//       var cur_key = data_keys[i]
//       col_el.node().innerHTML = cur_key
//
//     }
//
//     data.forEach(function (d) {
//       row = table_body.append("tr")
//       for (var i = 0; i < data_keys.length; i++) {
//         col_el = row.append("td")
//         var cur_key = data_keys[i]
//         var cur_val = d[cur_key]
//
//         if (typeofNumber(cur_val) == "float") {
//           cur_val = cur_val.slice(0, 7)
//         }
//         col_el.node().innerHTML = cur_val
//
//       }
//
//     })
//     update_table_header_width(chart_id)
//   })
// }


function classification_results_table(){ // this is the feature importance table
  var chart_id = "oasis_classification_feature_importance_table"


  // select the right data
  var first_sub_object, second_sub_object
  if (charts_config.document_state.balanced) {
    first_sub_object = "balanced"
  } else {
    first_sub_object = "unbalanced"
  }

  if (charts_config.document_state.optimized) {
    second_sub_object = "optimized_variables"
  } else {
    second_sub_object = "all_variables"
  }

  // sort the feature importance from ascending to decending
  var data = features[first_sub_object][second_sub_object]
  data.sort(function(a, b) {
    return parseFloat(a["Feature Importance"]) - parseFloat(b["Feature Importance"])
  })
  data.reverse()

  // create the feature importance table
  var table = d3.select("#" + chart_id)
  table.selectAll("*").remove()
  var table_head = table.append("thead") // the header (seperate from body to keep it always at the top)
  var table_body = table.append("tbody") // the body


  var data_keys = keys(data[0]) // feature importance keys
  var row, col_el // iterators

  // NEW mirna
  // add mirna data
  data_keys = data_keys.concat(['Predicted Targets', 'Known Targets', 'Select'])

  // populate the header
  row = table_head.append("tr")
  for (var i = 0; i < data_keys.length; i++) {
    col_el = row.append("th")
    var cur_key = data_keys[i]
    col_el.node().innerHTML = cur_key
    if (i==7) {
      col_el.on('mouseover', function(d, i) {
        d3.select(this).node().innerHTML = data_keys[5]
        d3.select(this).node().innerHTML += "<p style='font-size:10px'>miRTarBase & miRecords</p>"
      })
      col_el.on('mouseout', function(d, i){
        d3.select(this).node().innerHTML = data_keys[5]
      })
    }

    if (i==6) {
      col_el.on('mouseover', function(d, i) {
        d3.select(this).node().innerHTML = data_keys[6]
        d3.select(this).node().innerHTML += "<p style='font-size:10px'>miRanda</p>"
      })
      col_el.on('mouseout', function(d, i){
        d3.select(this).node().innerHTML = data_keys[6]
      })
    }

  }

  // iterator over the data and populate the table
  data.forEach(function (d, ind) {
    row = table_body.append("tr") // make the row
    // iterator over the header values
    for (var i = 0; i < data_keys.length; i++) {
      // if not the check box
      if (i != 8) {
        // fill with the value
        col_el = row.append("td")
        var cur_key = data_keys[i]
        var cur_val = d[cur_key]

        // extract the mirna info
        mirna = data[ind]['mature']
        // fix because Oasis truncates fourth values (e.g. -5p)
        mirna = mirna.split('-').slice(0,3).join('-')

        // custom options for pred / targets
        targets_flag = false // use to know whether to bind the pop up window
        targets = undefined
        if (cur_key.indexOf('Predicted Targets') != -1 || cur_key.indexOf('Known Targets')!= -1) {
          if (mirna_targets[mirna] == undefined) { // not a key, return N/a
            cur_val = 'N/a'
          } else {
            // grab the targets
            targets = mirna_targets[mirna][cur_key.toLowerCase().replace(' ', '_')]
            cur_val = targets.length
            targets_flag = true // flag to know to bind the cell to the pop up
          }
        }

        // some pre-processing of values
        if (typeofNumber(cur_val) == "float") {
          cur_val = cur_val.slice(0, 7) // only 7 characters (0, ., and 5 decimal points)
        }
        // set the text value
        col_el.node().innerHTML = cur_val

        // create pop-ups for targets
        if (targets_flag) {
          col_el.on('click', target_toggle)
        }

        // bind the open miRbase function to mature miRNAs
        if (cur_key == 'mature') {
          col_el.on('click', call_mirbase)
        }

        function call_mirbase(d, i , j) {
          // open mirbase if click is miRNA
            cell = d3.select(this)
            parent = d3.select(cell.node().parentNode)
            if (parent.select(':nth-child(2)').node().innerHTML == 'miRNA'){
                url = 'http://www.mirbase.org/cgi-bin/mirna_entry.pl?acc='
                url += cell.node().innerHTML
                window.open(url)
            }
        }

        // for toggling the target pop up box
        function target_toggle(d, i, j) {
          cell = d3.select(this)
          tog = eval(cell.attr('toggle'))
          mirna = d3.select(cell.node().parentNode).select(':nth-child(1)').node().innerHTML
          mirna = mirna.split('-').slice(0,3).join('-')

          if (d3.select(cell.node().parentNode).select(':nth-child(6)').node() == cell.node()) {
            cur_key = 'Predicted Targets'
          } else {
            cur_key = 'Known Targets'
          }

          cur_key = cur_key.toLowerCase().replace(' ', '_')


          if (!tog) {
            cell.attr('toggle', true)



            if (mirna_targets[mirna] != undefined) {
              targets = mirna_targets[mirna][cur_key]

              d3.selectAll('#target_div').remove()

              mouse = d3.mouse(d3.select('body').node())
              target_div = d3.select('body')
              .append('div').attr('class',cur_key + ' target_div').attr('id', 'target_div')
              .style('position', 'absolute')
              .style('left', mouse[0]+'px')
              .style('top', mouse[1]+'px')
              .style('background-color', '#E9EFFA')
              .style('border', '1px solid black')
              .style('border-radius', '0%')
              .style('height', '100px')
              .style('width', '150px')
              // .style('overflow-y', 'scroll')




              copyTextArea = target_div.append('textarea')
              .attr('class', 'targetList')
              .style('height', '95%')
              .style('background-color', '#E9EFFA')
              .style('width', '95%')
              .style('resize', 'none')
              .style('border', 'none')
              .style('display', 'inline-block')
              .style('position', 'relative')
              .style('top', '0%')
              // .style('padding', '2%')

              for (var i = 0; i < targets.length; i++) {
                copyTextArea.node().innerHTML+= targets[i] + '\n'
              }


              copyButton = target_div.append('button').attr('class', 'copyButton')
              .style('position', 'absolute')
              .style('left', '50%')
              .style('top', '0%')
              copyButton.node().innerHTML = 'copy'

              copyButton = document.querySelector('button.copyButton').addEventListener('click', function(event) {
                qs = document.querySelector('textarea.targetList')
                qs.select()

                try {
                  successfulCopy = document.execCommand('copy')
                  msg = successfulCopy ? 'yes' : 'no'
                } catch (err) {}
              })



              // target_div.append('ul').attr('class','targetList').selectAll('li').data(targets)
              // .enter().append('li').text(function(d, i){return d})
              // .style('list-style-type', 'none')
              // .attr('class', 'targets')
            }
          } else {
          cell.attr('toggle', false)
          d3.select('.'+cur_key).remove()
        }

      }

    } else {
      // this is the check box
      col_el = row.append("td")
      col_el.node().innerHTML = '<input type="checkbox">'

      // bind the on click select it
      col_el.on('click', mature_select)
      // starts not selected
      col_el.attr('selected', false)

      function mature_select(d, i , j) {
        cell = d3.select(this)
        sel = eval(cell.attr('selected'))
        if (!sel) {
          cell.style('background-color', '#03A9F4')
          cell.attr('selected', true)
          c = cell.attr('class')+' mature_selected'
          cell.attr('class', c)
          cell.select('input').property('checked', true)
        } else {
          cell.style('background-color', '')
          cell.attr('selected', false)
          c = cell.attr('class').replace('mature_selected', '')
          cell.attr('class', c)
          cell.select('input').property('checked', false)

        }
      }
    }
  }
  })
  update_table_header_width(chart_id)
}


function callEnrichmentAnalysis() {
  enrichment_analyses_to_run = [] // which apis to call
  mirna_to_run = [] // needed in case of species details
  targets_to_run = [] // targets to send to api

  // loop over api check boxes and store checked ones
  d3.selectAll('input.enrichment-selection').each(function(d, i) {
    node = d3.select(this)
    if (node.property('checked')) {
      enrichment_analyses_to_run.push(node.node().value)
    }
  })


  if (enrichment_analyses_to_run.length == 0) {
    // nothing to do, no reason to do more js
    alert('Please first select the enrichment services you want to use (GeneProfiler, Genemania, DAVID, ...).');
    return ""
  }

  //check if the user has selected any miRNAs
 if( document.getElementsByClassName('mature_selected').length == 0 ) {
	 alert('Please use the checkboxes to select microRNAs that will be used in the enrichment tests.');
	 return "";
 }

  // loop over selected mature miRNAs
  d3.selectAll('.mature_selected').each(function(d,i){
    // this is the check box cell
    cell = d3.select(this)
    // this is the table row
    parent = d3.select(cell.node().parentNode)
    // this is the text of the mature column of the current row
    selected_mirna = parent.select(':nth-child(1)').node().innerHTML

    // extract the targets
    cur_tars = get_targets_of_mirna(parent.select(':nth-child(1)'))
    mirna_to_run.push(selected_mirna)
    targets_to_run = targets_to_run.concat(cur_tars)
  })


  if (enrichment_analyses_to_run.indexOf("DAVID") != -1) {
    open_david(targets_to_run)
  }

  if (enrichment_analyses_to_run.indexOf("Genemania") != -1) {
    open_genemania(mirna_to_run, targets_to_run)
  }

  if (enrichment_analyses_to_run.indexOf("STRING") != -1) {
    open_string(targets_to_run)
  }

  if (enrichment_analyses_to_run.indexOf("STITCH") != -1) {
    open_stitch(targets_to_run)
  }


  if (enrichment_analyses_to_run.indexOf("gProfiler") != -1) {
    open_gprofiler(mirna_to_run, targets_to_run)
  }

}

function get_targets_of_mirna(cell) {
  // again, oasis only uses the first 3 specifiers and not the (-5p, etc)
  cur_mirna = cell.node().innerHTML.split('-').slice(0,3).join('-')
  // only return targets for miRNA
  type = d3.select(cell.node().parentNode).select(':nth-child(2)').node().innerHTML
  if (type != 'miRNA'){ return[] }
  // extract targets from json
  known_tar = mirna_targets[cur_mirna]["known_targets"]
  return known_tar
}

function open_david(targets_to_run) {
  david_url = 'http://david.abcc.ncifcrf.gov/api.jsp?type=OFFICIAL_GENE_SYMBOL&ids='
  suffix = '&tool=summary'
  api = ''
  for (var i = 0; i < targets_to_run.length; i++) {
    cur_target = targets_to_run[i]
    api += cur_target
    if (i != targets_to_run.length - 1) {
      api += ','
    }

  }
  url = david_url + api + suffix
  win = window.open(url)
  // win.focus()
}


species_ids = {
  'hsa': {'id': '9606','species': 'homo_sapiens'},
  'mmu': {'id': '10090','species': 'mus_musculus'},
  'dme': {'id': '7227','species': 'drosophila_melanogaster'},
  'cel': {'id': '6239','species': 'caenorhabditis_elegans'},
  'dre': {'id': '7955','species': 'danio_rerio'}
}

function open_genemania(mirna_to_run, targets_to_run) {
  species = ''
  for (var i = 0; i < mirna_to_run.length; i++) {
    if(mirna_to_run[i].indexOf('-') != -1){
      species = mirna_to_run[i].split('-')[0]
      if (d3.keys(species_ids).indexOf(species) != -1) {
        species = species_ids[species]['species']
        break
      }
    }
  }

  genemania_url = 'http://genemania.org/search/'

  api = ''
  for (var i = 0; i < targets_to_run.length; i++) {
    cur_tar = targets_to_run[i]
    api += cur_tar

    if (i != targets_to_run.length - 1) {
      api += '|'
    }

  }
  url = genemania_url +species+'/'+ api
  win = window.open(url)
}

function open_gprofiler(mirna_to_run, targets_to_run) {
  species = ''
  score = '0'

  for (var i = 0; i < mirna_to_run.length; i++) {
    if(mirna_to_run[i].indexOf('-') != -1){
      species = mirna_to_run[i].split('-')[0]
      if (d3.keys(species_ids).indexOf(species) != -1) {
        [species_genus, species_name] = species_ids[species]['species'].split('_')
        species = species_genus[0] + species_name
        break
      }
    }
  }


  gprofiler_url = 'http://biit.cs.ut.ee/gprofiler/gcocoa.cgi?organism='
  api = ''
  for (var i = 0; i < targets_to_run.length; i++) {
    cur_tar = targets_to_run[i]
    api += cur_tar
    if (i != targets_to_run.length - 1) {
      api += '%0A'
    }
  }

  url = gprofiler_url + species + '&output=html&significant=1&ordered_query=1&query=' + api
  win = window.open(url)

}

function open_string(targets_to_run) {
  species = ''
  score = '0'
  for (var i = 0; i < mirna_to_run.length; i++) {
    if(mirna_to_run[i].indexOf('-') != -1){
      species = mirna_to_run[i].split('-')[0]
      if (d3.keys(species_ids).indexOf(species) != -1) {
        species = species_ids[species]['id']
        break
      }
    }
  }

  string_url = 'http://string-db.org/api/image/networkList?identifiers='
  api = ''
  suffix = '&limit=10&network_flavor=evidence'
  for (var i = 0; i < targets_to_run.length; i++) {
    cur_tar = targets_to_run[i]
    api += cur_tar

    if (i != targets_to_run.length - 1) {
      api += '%0D'
    }

  }

  url = string_url + api + '&species=' + species + '&required_score=950' + suffix
  win = window.open(url)
}

function open_stitch(targets_to_run) {
  species = ''
  score = '0'
  for (var i = 0; i < mirna_to_run.length; i++) {
    if(mirna_to_run[i].indexOf('-') != -1){
      species = mirna_to_run[i].split('-')[0]
      if (d3.keys(species_ids).indexOf(species) != -1) {
        species = species_ids[species]['id']
        break
      }
    }
  }

  string_url = 'http://stitch.embl.de/api/image/networkList?identifiers='
  api = ''
  suffix = '&limit=10&network_flavor=evidence'
  for (var i = 0; i < targets_to_run.length; i++) {
    cur_tar = targets_to_run[i]
    api += cur_tar

    if (i != targets_to_run.length - 1) {
      api += '%0D'
    }

  }

  url = string_url + api + '&species=' + species + '&required_score=950' + suffix
  win = window.open(url)
}


function update_table_header_width(table_id) {
  var table_head = d3.select("table#" + table_id).select("thead")
  var table_body = d3.select("table#" + table_id).select("tbody")


  var table_body_row = table_body.select("tr")
  var table_body_row_elements = table_body_row.selectAll("td")

  var column_widths = []
  table_body_row_elements.each(function() {
    column_widths.push(d3.select(this).style("width"))
  })

  table_head.select("tr").selectAll("th").each(
    function(d, i) {
      d3.select(this).style("width", column_widths[i])
    })
}

function update_classification_results_table_download_link(id) {
  d3.select("#" + id).node()

  var link_to_xls = "ClassificationResults/RandomForest/"
  if (charts_config.document_state.balanced) {
    link_to_xls += "balanced/"
  } else {
    link_to_xls += "unbalanced/"
  }
  if (charts_config.document_state.optimized) {
    link_to_xls += "optimized_variables/"
  } else {
    link_to_xls += "all_variables/"
  }
  link_to_xls += "classification.xls"
  d3.select("#" + id).node().href = link_to_xls
}

function current_model_folder() {
  var model_folder = "ClassificationResults/RandomForest/"
  if (charts_config.document_state.balanced) {
    model_folder += "balanced/"
  } else {
    model_folder += "unbalanced/"
  }
  if (charts_config.document_state.optimized) {
    model_folder += "optimized_variables/"
  } else {
    model_folder += "all_variables/"
  }
  return model_folder
}

function update_download_attribute(id_of_object, file) {
  var model_folder = current_model_folder()
  var path_to_file = model_folder + file
  d3.select("#" + id_of_object).node().href = path_to_file
}







////////////NEW//////
// function make_fsn_plot(models) {
//   var svg, chart_group, data_extent, canvas, margins, max_draw_space, doc_state
//   var plot = "feature_selection_new"
//   // extract current model
//   var model = extractModel(models)
//   // extract corresponding data
//   var data = model[plot]
//
//   var quantile_keys = ["0.00", "0.25", "0.50", "0.75", "1.00"]
//   var quantile_data = extract_quantile_data(data, "number_of_features", quantile_keys)
//   // Order in which lines are added (necessary if having splines filled with colors)
//   var quantile_keys = ["Q0", "Q4", "Q1", "Q3", "Q2"]
//
//
//   svg = make_chart_svg(plot)
//   if (svg.select("g.plot_group").empty()) {
//     chart_group = svg.append("g").attr("class", "plot_group")
//   } else {
//     chart_group = svg.select("g.plot_group")
//   }
//   chart_group.data(data)
//
//   var data_extent, x_axis_label, y_axis_label
//   data_extent = {
//     "max": {
//       "x": d3.max(quantile_data.Q4, function(d) {return d.x}),
//       "y": d3.max(quantile_data.Q4, function(d) {return d.y})
//     },
//     "min": {
//       "x": d3.min(quantile_data.Q0, function(d) {return d.x}),
//       "y": d3.min(quantile_data.Q0, function(d) {return d.y})
//     }
//   }
//
//
//   canvas = extract_canvas_from_svg(svg)
//   margins = make_margins(chart_group, canvas, data_extent, true)
//   max_draw_space = calculate_maximum_drawing_space(canvas, margins)
//   make_title(chart_group, ["Feature Selection New"], margins, canvas, max_draw_space)
//
//
//   x_axis_label = "Features"
//   y_axis_label = "Cross-validated error"
//   var x_scale = d3.scaleLinear().domain([data_extent.min.x, data_extent.max.x]).range([0, max_draw_space.x])
//   var y_scale = d3.scaleLinear().domain([data_extent.max.y, data_extent.min.y]).range([0, max_draw_space.y])
//   make_axes(chart_group, x_scale, y_scale, canvas, margins, max_draw_space, x_axis_label, y_axis_label)
//
//   var quantile_line = d3.line()
//                         .x(function(d) {return x_scale(d.x)})
//                         .y(function(d) {return y_scale(d.y)})
//                         .curve(d3.curveBasis)
//
//   // Handle Lines
//   if (chart_group.selectAll("g." + plot + "_quantile_lines_group").empty()) {
//     var quantile_lines = chart_group.append("g")
//     .attr("class", plot + "_quantile_lines_group")
//     .selectAll("g." + plot + "_quantile_line_group")
//     .data(quantile_keys)
//     .enter()
//     .append("g")
//     .attr("class", plot + "_quantile_line_group")
//     .each(function(d, i) {
//       var quantile_path = d3.select(this).append("path")
//       if (d == "Q0" || d == "Q4") { // outer most lines
//         o = d == "Q0" ? "Q1" : "Q3"
//         quantile_path.datum(quantile_data[d].concat(quantile_data[o].reverse()))
//         .attr("fill", charts_config.plots.quantile_spline.outer_fill_color)
//                       .attr("stroke", charts_config.plots.quantile_spline.outer_line_color)
//                       .attr("stroke-width", charts_config.plots.quantile_spline.stroke_width)
//                       .style("opacity", charts_config.plots.quantile_spline.opacity)
//       } else if (d == "Q1" || d == "Q3") { // inner lines
//         quantile_path.datum(quantile_data[d].concat(quantile_data["Q2"]))
//         .attr("fill", charts_config.plots.quantile_spline.inner_fill_color)
//                       .attr("stroke", charts_config.plots.quantile_spline.inner_line_color)
//                       .attr("stroke-width", charts_config.plots.quantile_spline.stroke_width)
//                       .style("opacity", charts_config.plots.quantile_spline.opacity)
//       } else { // median line
//         quantile_path.datum(quantile_data[d])
//         .attr("fill", charts_config.plots.quantile_spline.median_fill_color)
//         .attr("stroke", charts_config.plots.quantile_spline.median_line_color)
//         .attr("stroke-width", charts_config.plots.quantile_spline.median_stroke_width)
//         .style("opacity", charts_config.plots.quantile_spline.median_opacity)
//       }
//       quantile_path.attr("d", quantile_line).attr("transform",
//       "translate(" + (margins.x.left + margins.axes.y) + "," + (margins.y.top + margins.title + margins.buttons) + ")")
//     })
//   }
//
//   // Handle Points
//   var previous_points = chart_group.selectAll("g." + plot + "_hidden_points_group").size()
//   var current_points = quantile_keys.map(function(d) {return quantile_data[d].length}).reduce(function(a, b) { return a + b; }, 0)
//
//   if (previous_points < current_points) {
//
//   }
//
//
//
// }
//
// function extract_quantile_data(list_of_quantile_objects, x_key, quantile_keys) {
//   var quantile_data = {"Q0":[],"Q1":[],"Q2":[],"Q3":[],"Q4":[]}
//
//   // For each x coordinate
//   for (var i = 0; i < list_of_quantile_objects.length; i++) {
//     var current_quantile_object = list_of_quantile_objects[i]
//     // For each quantile extract x and y coordinates
//     for (var j = 0; j < quantile_keys.length; j++) {
//       var current_quantile_data = {}
//       var cur_q_key = quantile_keys[j]
//       current_quantile_data.x = current_quantile_object[x_key]
//       current_quantile_data.y = current_quantile_object.quantiles[cur_q_key]
//       // console.log(i, j, cur_q_key, current_quantile_data.y)
//       quantile_data[("Q" + j)].push(current_quantile_data)
//     }
//
//   }
//   return quantile_data
// }
